#ifndef RO_H
#define RO_H

#include <stdexcept>
#include <string>
#include <memory>
#include <functional>
#include <tuple>

#define None(...) ro::Option<__VA_ARGS__>()
#define Some(...) ro::Option(__VA_ARGS__)
#define ERR(fmt, ...) Error(ro::sfmt("%s:%d: " fmt, __FILE__, __LINE__, ##__VA_ARGS__))
#define try_unwrap(l, ...)          \
    try                             \
    {                               \
        l = (__VA_ARGS__).unwrap(); \
    }                               \
    catch (ro::Error e)             \
    {                               \
        return e;                   \
    }

#define ERR_INVALID_ACESS "Invalid access"

namespace ro
{
    template <typename... Args>
    std::string sfmt(const std::string &format, Args... args)
    {
        int size_s = std::snprintf(nullptr, 0, format.c_str(), args...) + 1; // Extra space for '\0'
        if (size_s <= 0)
        {
            throw std::runtime_error("Error during formatting string.");
        }
        auto size = static_cast<size_t>(size_s);
        std::unique_ptr<char[]> buf(new char[size]);
        std::snprintf(buf.get(), size, format.c_str(), args...);
        return std::string(buf.get(), buf.get() + size - 1); // We don't want the '\0' inside
    }

    class Error : public std::runtime_error
    {
    public:
        Error(const std::string &s) : runtime_error(s){};
        Error(const char *s) : runtime_error(s){};
        Error() : runtime_error(""){};

        inline friend bool operator==(const Error &l, const Error &r)
        {
            return std::string(l.what()) == std::string(r.what());
        }

        friend inline Error operator+(const Error &l, const Error &r)
        {
            return Error(std::string(l.what()) + "\n" + r.what());
        }
    };

    template <typename T, typename E = ro::Error>
    class Result;

    template <typename T>
    class Option
    {

    public:
        using type_t = T;
        Option(T obj) : m_some(obj), m_notnull(true){};
        Option() : m_notnull(false){};

        Option<T> &operator=(const T &value)
        {
            insert(value);
            return *this;
        };

        inline friend bool operator==(const Option<T> &l, const Option<T> &r)
        {
            if (!l.m_notnull && !r.m_notnull)
            {
                return true;
            }
            if (l.m_notnull && r.m_notnull)
            {
                return l.m_some == r.m_some;
            }
            return false;
        };

        inline friend Option<T> operator|(const Option<T> &l, const Option<T> &r)
        {
            if (l.is_some())
            {
                return l;
            }
            return r;
        }

        template <typename U>
        inline friend Option<U> operator&(const Option<T> &l, const Option<U> &r)
        {
            if (l.is_some())
            {
                return r;
            }
            return Option<U>();
        }

        inline friend Option<T> operator^(const Option<T> &l, const Option<T> &r)
        {
            if (l.is_some() != r.is_some())
            {
                if (l.is_some())
                {
                    return l;
                }
                return r;
            }
            return Option<T>();
        }

        inline bool is_none() const { return !m_notnull; };
        inline bool is_some() const { return m_notnull; };

        template <typename U>
        inline U match(const std::function<U(T &)> &expected, const std::function<U()> &unexpected)
        {
            if (is_some())
            {
                return expected(m_some);
            }
            return unexpected();
        }

        template <typename U>
        inline U match(const std::function<U(const T &)> &expected, const std::function<U()> &unexpected) const
        {
            if (is_some())
            {
                return expected(m_some);
            }
            return unexpected();
        }

        inline const T &unwrap() const
        {
            if (m_notnull)
            {
                return m_some;
            }
            throw Error(ERR_INVALID_ACESS);
        }

        inline T &unwrap()
        {
            if (m_notnull)
            {
                return m_some;
            }
            throw Error(ERR_INVALID_ACESS);
        }

        inline const T &unwrap_or(const T &alt) const
        {
            if (m_notnull)
            {
                return m_some;
            }
            return alt;
        }

        inline T unwrap_or_else(const std::function<T()> &fn) const
        {
            if (m_notnull)
            {
                return m_some;
            }
            return fn();
        }

        inline T &get_or_insert(const T &value)
        {
            if (!m_notnull)
            {
                insert(value);
            }
            return m_some;
        }

        inline T &insert(const T &value)
        {
            m_some = value;
            m_notnull = true;
            return m_some;
        }

        template <typename E>
        inline Result<T, E> ok_or(const E &err)
        {
            if (m_notnull)
            {
                return Result<T, E>(m_some);
            }
            return Result<T, E>(err);
        }

        template <typename U>
        inline Option<U> and_then(const std::function<Option<U>(const T &)> &fn)
        {
            if (m_notnull)
            {
                return fn(m_some);
            }
            return Option<U>();
        }

        inline Option<T> or_else(const std::function<Option<T>()> &fn)
        {
            if (m_notnull)
            {
                return *this;
            }
            return fn();
        }

        inline bool into(T &output) const
        {
            if (is_some())
            {
                output = m_some;
                return true;
            }
            return false;
        }

        template <typename U>
        inline Option<U> map(const std::function<U(const T &)> &fn)
        {
            if (is_some())
            {
                return fn(m_some);
            }
            return Option<U>();
        }

        template <typename U>
        inline U map_or(const U &defv, const std::function<U(const T &)> &fn)
        {
            if (is_some())
            {
                return fn(m_some);
            }
            return defv;
        }

        inline Option<T> filter(const std::function<bool(const T &)> &fn)
        {
            if (is_some() && fn(m_some))
            {
                return Option<T>(m_some);
            }
            return Option<T>();
        }

        inline Option<T> take()
        {
            if (m_notnull)
            {
                m_notnull = false;
                return Option<T>(m_some);
            }
            return Option<T>();
        }

        inline Option<T> take_if(const std::function<bool(const T &)> &fn)
        {
            if (m_notnull && fn(m_some))
            {
                return take();
            }
            return Option<T>();
        }

        template <typename U>
        inline Option<std::tuple<T, U>> zip(const Option<U> &other)
        {
            if (m_notnull && other.is_some())
            {
                return Option<std::tuple<T, U>>(std::tuple<T, U>(m_some, other.unwrap()));
            }
            return Option<std::tuple<T, U>>();
        }

        template <typename U, typename R>
        inline Option<R> zip_with(const Option<U> &other, const std::function<R(const T &, const U &)> &fn)
        {
            if (m_notnull && other.is_some())
            {
                return fn(m_some, other.unwrap());
            }
            return Option<R>();
        }

        template <typename Y, typename Z>
        inline typename std::enable_if<std::is_base_of<std::tuple<Y, Z>, T>::value, std::tuple<Option<Y>, Option<Z>>>::type unzip()
        {
            if (is_some())
            {
                return std::tuple<Option<Y>, Option<Z>>(Some(std::get<Y>(m_some)), Some(std::get<Z>(m_some)));
            }
            return std::tuple<Option<Y>, Option<Z>>(None(Y), None(Z));
        }

        template <typename X = T>
        inline typename std::enable_if<std::is_base_of<Option<typename X::type_t>, X>::value, X>::type flatten()
        {
            if (is_some())
            {
                return m_some;
            }
            return None(typename X::type_t);
        }

        template <typename X = T>
        inline typename std::enable_if<std::is_base_of<Result<typename X::type_ok_t, typename X::type_err_t>, X>::value, Result<Option<typename X::type_ok_t>, typename X::type_err_t>>::type transpose()
        {
            if (is_none())
            {
                return Option<typename X::type_ok_t>();
            }
            if (m_some.is_ok())
            {
                return m_some.ok();
            }
            return m_some.unwrap_err();
        }

        inline Option<T> replace(const T &value)
        {

            Option<T> ret;
            if (is_some())
            {
                ret = m_some;
            }
            insert(value);
            return ret;
        }

    private:
        T m_some;
        bool m_notnull;
    };

    template <typename T, typename E>
    class Result
    {

    public:
        using type_ok_t = T;
        using type_err_t = E;
        Result(T obj) : m_result(obj){};
        Result(E err) { m_error.insert(err); };
        Result(){};

        inline Result<T, E> &operator=(const T &value)
        {
            m_result = value;
            m_error = None(E);
            return *this;
        };
        inline Result<T, E> &operator=(const E &err)
        {
            m_error.insert(err);
            return *this;
        };

        inline friend bool operator==(const Result<T, E> &l, const Result<T, E> &r)
        {
            if (l.is_ok() != r.is_ok())
            {
                return false;
            }
            if (l.is_ok())
            {
                return l.m_result == r.m_result;
            }
            return l.m_error == r.m_error;
        };

        inline friend bool operator==(const Result<T, E> &l, const E &r)
        {
            if (l.is_ok())
            {
                return false;
            }
            return l.m_error == r;
        };

        inline friend bool operator==(const Result<T, E> &l, const T &r)
        {
            if (l.is_err())
            {
                return false;
            }
            return l.m_result == r;
        };

        inline friend Result<T, E> operator|(const Result<T, E> &l, const Result<T, E> &r)
        {
            if (l.is_ok())
            {
                return l;
            }
            return r;
        }

        template <typename U>
        inline friend Result<U, E> operator&(const Result<T, E> &l, const Result<U, E> &r)
        {
            if (l.is_ok())
            {
                return r;
            }
            return Result<U, E>(l.m_error.unwrap());
        }

        inline bool is_ok() const { return m_error.is_none(); };
        inline bool is_err() const { return m_error.is_some(); };

        template <typename U>
        inline U match(const std::function<U(T &)> &expected, const std::function<U(E &)> &unexpected)
        {
            if (is_ok())
            {
                return expected(m_result);
            }
            return unexpected(m_error.unwrap());
        }

        template <typename U>
        inline U match(const std::function<U(const T &)> &expected, const std::function<U(const E &)> &unexpected) const
        {
            if (is_ok())
            {
                return expected(m_result);
            }
            return unexpected(m_error.unwrap());
        }

        template <typename F>
        inline Result<T, F> or_else(const std::function<Result<T, F>(const E &)> &fn)
        {
            if (is_ok())
            {
                return Result<T, F>(m_result);
            }
            return fn(m_error.unwrap());
        }
        inline const T &unwrap() const
        {
            if (is_ok())
            {
                return m_result;
            }
            throw m_error.unwrap();
        }

        inline T &unwrap()
        {
            if (is_ok())
            {
                return m_result;
            }
            throw m_error.unwrap();
        }

        inline const E &unwrap_err() const
        {
            if (is_err())
            {
                return m_error.unwrap();
            }
            throw m_result;
        }

        inline const T &expect(const E &e) const
        {
            if (is_ok())
            {
                return m_result;
            }
            throw e + m_error.unwrap();
        }

        template <typename U>
        inline Result<U, E> map(const std::function<U(const T &)> &fn)
        {
            if (is_ok())
            {
                return fn(m_result);
            }
            return m_error.unwrap();
        }

        template <typename E1>
        inline Result<T, E1> map_err(const std::function<E1(const E &)> &fn)
        {
            if (!is_ok())
            {
                return fn(m_error.unwrap());
            }
            return m_result;
        }

        template <typename U>
        inline U map_or(const U &defv, const std::function<U(const T &)> &fn)
        {
            if (is_ok())
            {
                return fn(m_result);
            }
            return defv;
        }

        template <typename U>
        inline Result<U, E> and_then(const std::function<Result<U, E>(const T &)> &fn)
        {
            if (is_ok())
            {
                return fn(m_result);
            }
            return Result<U, E>(m_error.unwrap());
        }

        inline const T &unwrap_or(const T &alt) const
        {
            if (is_ok())
            {
                return m_result;
            }
            return alt;
        }

        inline bool into(T &output) const
        {
            if (is_ok())
            {
                output = m_result;
                return true;
            }
            return false;
        }

        inline T unwrap_or_else(const std::function<T(const E &)> &fn)
        {
            if (is_ok())
            {
                return m_result;
            }
            return fn(m_error.unwrap());
        }

        template <typename X = T>
        inline typename std::enable_if<std::is_base_of<Result<typename X::type_ok_t, typename X::type_err_t>, X>::value, X>::type flatten()
        {
            if (is_ok())
            {
                return m_result;
            }
            return m_error.unwrap();
        }

        template <typename X = T>
        inline typename std::enable_if<std::is_base_of<Option<typename X::type_t>, X>::value, Option<Result<typename X::type_t, E>>>::type transpose()
        {
            if (is_err())
            {
                return Result<typename X::type_t, E>(m_error.unwrap());
            }

            if (m_result.is_none())
            {
                return None(Result<typename X::type_t, E>);
            }

            return Result<typename X::type_t, E>(m_result.unwrap());
        }

        inline const Option<E> &err() const { return m_error; }
        inline Option<T> ok() const { return is_ok() ? Option<T>(m_result) : Option<T>(); }

    private:
        T m_result;
        Option<E> m_error;
    };
}

#endif